import 'dart:async';

import 'package:equatable/equatable.dart';
import 'package:osam/osam.dart';
import 'package:osam/src/util/value_stream.dart';

typedef ValueMapper<ST, V> = V Function(ST state);
typedef StateMapper<ST> = void Function(ST state);

// ignore: must_be_immutable
abstract class BaseState<ST> extends Equatable {
  StreamController<ST> _stateBroadcaster;
  int _lastKnownHashCode = -1;

  void _initController() {
    _stateBroadcaster = StreamController<ST>.broadcast(onCancel: () {
      _stateBroadcaster.close();
    });
  }

  ValueStream<ST> get _stateStream => ValueStream(_stateBroadcaster.stream.map((state) => state), this as ST);

  ValueStream<V> propertyStream<V>(ValueMapper<ST, V> mapper) {
    if (_stateBroadcaster == null || _stateBroadcaster.isClosed) _initController();
    return _stateStream.map((event) => mapper(event));
  }

  void update(StateMapper<ST> change) {
    if (_stateBroadcaster == null || _stateBroadcaster.isClosed) _initController();
    _lastKnownHashCode = hashCode;
    try {
      change(this as ST);
    } catch (e) {
      handleError(e);
    }
    if (_lastKnownHashCode != hashCode) {
      _stateBroadcaster.sink.add(this as ST);
    }
  }

  void handleError(Object error) => print(error);

  @override
  List<Object> get props;
}

class ComparableWrapper<V> extends Equatable {
  final V _value;

  ComparableWrapper(this._value);

  @override
  List<Object> get props => [_value];
}
